import logging
import sys
from pathlib import Path
from typing import Optional

from nicegui import ui

from .. import rosys
from ..event import Event
from .communication import Communication
from .lizard_firmware import LizardFirmware


class RobotBrain:
    """This module manages the communication with a [Zauberzeug Robot Brain](https://zauberzeug.com/robot-brain.html).

    It expects a communication object, which is used for the actual read and write operations.
    Besides providing some basic methods like configuring or restarting the microcontroller, it augments and verifies checksums for each message.
    """

    def __init__(self, communication: Communication) -> None:
        self.LINE_RECEIVED = Event()
        """a line has been received from the microcontroller (argument: line as string)"""

        self.log = logging.getLogger('rosys.robot_rain')

        self.communication = communication
        self.lizard_code = ''
        self.lizard_firmware = LizardFirmware(self)

        self.waiting_list: dict[str, Optional[str]] = {}
        self.clock_offset: Optional[float] = None
        self.hardware_time: Optional[float] = None

        rosys.on_startup(self.enable_esp)

    def developer_ui(self) -> None:
        async def online_update() -> None:
            await self.lizard_firmware.download()
            await self.lizard_firmware.flash_core()
            await self.restart()
            await self.configure()
            await self.lizard_firmware.flash_p0()

        async def local_update() -> None:
            await self.lizard_firmware.flash_core()
            await self.restart()
            await self.configure()
            await self.lizard_firmware.flash_p0()

        with ui.row().classes('items-center'):
            ui.label().bind_text_from(self.lizard_firmware, 'online_version', backward=lambda x: f'Online: {x or "?"}')
            online_update_button = ui.button(on_click=online_update).props('icon=file_download flat round dense') \
                .tooltip('Download and flash online version to Core and P0 microcontrollers')
        with ui.row().classes('items-center'):
            ui.label().bind_text_from(self.lizard_firmware, 'local_version', backward=lambda x: f'Local: {x or "?"}')
            local_update_button = ui.button(on_click=local_update).props('icon=file_download flat round dense') \
                .tooltip('Flash local version to Core and P0 microcontrollers')
        with ui.row().classes('items-center'):
            ui.label().bind_text_from(self.lizard_firmware, 'core_version', backward=lambda x: f'Core: {x or "?"}')
            configure_button = ui.button(on_click=self.configure).props('icon=build flat round dense') \
                .tooltip('Configure microcontrollers')
        with ui.row().classes('items-center'):
            ui.label().bind_text_from(self.lizard_firmware, 'p0_version', backward=lambda x: f'P0: {x or "?"}')

        def update_visibility() -> None:
            online_update_button.visible = \
                self.lizard_firmware.online_version != self.lizard_firmware.core_version or \
                self.lizard_firmware.online_version != self.lizard_firmware.p0_version
            local_update_button.visible = \
                self.lizard_firmware.local_version != self.lizard_firmware.core_version or \
                self.lizard_firmware.local_version != self.lizard_firmware.p0_version
            configure_button.visible = \
                self.lizard_firmware.local_checksum != self.lizard_firmware.core_checksum
        ui.timer(1.0, update_visibility)

        with ui.row().classes('items-center'):
            ui.button('Refresh', on_click=self.lizard_firmware.read_all).props('outline') \
                .tooltip('Read installed and available versions')
            with ui.row():
                with ui.menu() as menu:
                    ui.menu_item('Download', on_click=self.lizard_firmware.download) \
                        .tooltip('Download the latest Lizard firmware from GitHub')
                    ui.menu_item('Flash Core', on_click=self.lizard_firmware.flash_core) \
                        .tooltip('Flash the downloaded Lizard firmware to the Core microcontroller')
                    ui.menu_item('Flash P0', on_click=self.lizard_firmware.flash_p0) \
                        .tooltip('Flash the downloaded Lizard firmware to the P0 microcontroller')
                    ui.menu_item('Enable', on_click=self.enable_esp) \
                        .tooltip('Enable the microcontroller module (will later be done automatically)')
                    ui.menu_item('Configure', on_click=self.configure) \
                        .tooltip('Configure the microcontroller with the Lizard startup file')
                    ui.menu_item('Restart', on_click=self.restart) \
                        .tooltip('Restart the microcontroller')
                ui.button(on_click=menu.open).props('icon=more_vert flat round')

        ui.label().bind_text_from(self, 'clock_offset', lambda offset: f'Clock offset: {offset or 0:.3f} s')

    async def configure(self) -> None:
        rosys.notify('Configuring Lizard...')
        await self.send(f'!-')
        for line in self.lizard_code.splitlines():
            await self.send(f'!+{line}')
        await self.send(f'!.')
        await self.restart()
        rosys.notify('Lizard configured successfully.', 'positive')

    async def restart(self) -> None:
        await self.send(f'core.restart()')

    async def read_lines(self) -> list[tuple[float, str]]:
        lines: list[tuple[float, str]] = []
        millis = None
        while True:
            unchecked = await self.communication.read()
            line = check(unchecked)
            if not line:
                break
            words = line.split()
            if not words:
                continue
            first = words.pop(0)
            if first in self.waiting_list:
                self.waiting_list[first] = line
            if first == 'core':
                millis = float(words.pop(0))
                if self.clock_offset is None:
                    continue
                self.hardware_time = millis / 1000 + self.clock_offset
            self.LINE_RECEIVED.emit(line)
            lines.append((self.hardware_time, line))
        if millis is not None:
            self.clock_offset = rosys.time() - millis / 1000
        return lines

    async def send(self, msg: str) -> None:
        await self.communication.send(augment(msg))

    async def send_and_await(self, msg: str, ack: str, *, timeout: float = float('inf')) -> Optional[str]:
        self.waiting_list[ack] = None
        await self.send(msg)
        t0 = rosys.time()
        while self.waiting_list.get(ack) is None and rosys.time() < t0 + timeout:
            await rosys.sleep(0.1)
        return self.waiting_list.pop(ack) if ack in self.waiting_list else None

    def enable_esp(self) -> None:
        try:
            sys.path.insert(1, str(Path('~/.lizard').expanduser()))
            from esp import Esp
            params = self.lizard_firmware.flash_params
            devices = [param for param in params if param.startswith('/dev/')]
            esp = Esp(nand='nand' in params, xavier='xavier' in params, device=devices[0] if devices else None)
            with esp.pin_config():
                esp.activate()
        except Exception:
            self.log.exception('Could not enable ESP')

    def __del__(self) -> None:
        self.communication.disconnect()

    def __repr__(self) -> str:
        return f'<RobotBrain {self.communication}>'


def augment(line: str) -> str:
    checksum = 0
    for c in line:
        checksum ^= ord(c)
    return f'{line}@{checksum:02x}'


def check(line: Optional[str]) -> str:
    if line is None:
        return ""
    if line[-3:-2] == '@':
        check = int(line[-2:], 16)
        line = line[:-3]
        checksum = 0
        for c in line:
            checksum ^= ord(c)
        if checksum != check:
            return None
    return line



from itertools import groupby

import humanize
from matplotlib.colors import hsv_to_rgb, to_hex
from nicegui import ui

from .kpi_buckets import Month, TimeBucket, Week
from .kpi_logger import KpiLogger, str_to_date


def kpi_page(kpi_logger: KpiLogger,
             *,
             timespans: dict[int, str],
             positives: dict[str, str],
             negatives: dict[str, str],
             title: str = '',
             positive_title: str = '',
             negative_title: str = '') -> None:

    humanize.activate('de')

    @ui.page('/kpis')
    def page():
        with ui.row().style('margin:1em'):
            ui.markdown(f'### {title}').style('margin:1.5em;margin-top:-1.2em;')
            toggle = ui.toggle(timespans, value=list(timespans.keys())[0], on_change=lambda e: show(e.value))
        with ui.row():
            positive_chart = ui.chart({
                'title': {'text': positive_title},
                'chart': {'type': 'column'},
                'xAxis': {},
                'yAxis': {'title': False},
                'credits': False,
                'exporting': False,
            })
            negative_chart = ui.chart({
                'title': {'text': negative_title},
                'chart': {'type': 'column'},
                'plotOptions': {'series': {'stacking': 'normal'}},
                'xAxis': {},
                'yAxis': {'title': False},
                'colors': [to_hex(hsv_to_rgb((0, 1, i / 100.0))) for i in reversed(range(20, 120, 20))],
                'credits': False,
                'exporting': False,
            })
        ui.timer(5, lambda: show(toggle.value))

        def show(num_days: int) -> None:
            if num_days <= 7:
                time_buckets = kpi_logger.days[-num_days:]
            elif num_days % 7 == 0:
                weeks = [list(v) for _, v in groupby(kpi_logger.days,
                                                     key=lambda d: str_to_date(d.date).strftime('%y-%W'))][-num_days//7:]
                time_buckets = [Week.from_buckets(days_in_week) for days_in_week in weeks]
            elif num_days % 30 == 0 and num_days <= 90:
                months = [list(v) for _, v in groupby(kpi_logger.days,
                                                      key=lambda d: str_to_date(d.date).strftime('%y-%m'))][-num_days//30:]
                time_buckets = [Month.from_buckets(days_in_month) for days_in_month in months]
            else:
                raise ValueError(f'Unsupported number of days: {num_days}')

            keys = set(key for day in time_buckets for key in day.incidents if key in positives)
            data = {positives[key]: [day.incidents.get(key, 0) for day in time_buckets] for key in keys}
            positive_chart.options['xAxis']['categories'] = [_label(b) for b in time_buckets]
            positive_chart.options['series'] = [{'name': k, 'data': v} for k, v in sorted(data.items())]
            positive_chart.update()

            keys = set(key for day in time_buckets for key in day.incidents if key in negatives)
            data = {negatives[key]: [day.incidents.get(key, 0) for day in time_buckets] for key in keys}
            negative_chart.options['xAxis']['categories'] = [_label(b) for b in time_buckets]
            negative_chart.options['series'] = [{'name': k, 'data': v} for k, v in sorted(data.items())]
            negative_chart.update()


def _label(time_bucket: TimeBucket) -> str:
    if isinstance(time_bucket, Month):
        return str_to_date(time_bucket.date).strftime('%b')
    elif isinstance(time_bucket, Week):
        return f'#{str_to_date(time_bucket.date).isocalendar()[1]}'
    else:
        return humanize.naturaldate(str_to_date(time_bucket.date))



import asyncio
import logging
import os
import shlex
import signal
import subprocess
import uuid
from asyncio.subprocess import Process
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor
from concurrent.futures.process import BrokenProcessPool
from contextlib import contextmanager
from functools import partial, wraps
from typing import Callable, Generator, Optional

from psutil import Popen

from . import rosys

process_pool = ProcessPoolExecutor()
thread_pool = ThreadPoolExecutor(thread_name_prefix='run.py thread_pool')
running_cpu_bound_processes: list[int] = []  # NOTE is used in tests to advance time slower until computation is done
running_sh_processes: list[Process] = []
log = logging.getLogger('rosys.run')


async def io_bound(callback: Callable, *args: any, **kwargs: any):
    if rosys.is_stopping():
        return
    loop = asyncio.get_running_loop()
    try:
        return await loop.run_in_executor(thread_pool, partial(callback, *args, **kwargs))
    except RuntimeError as e:
        if 'cannot schedule new futures after shutdown' not in str(e):
            raise
    except asyncio.exceptions.CancelledError:
        pass


def awaitable(func: Callable) -> Callable:
    """decorator to wrap a normal function into an asyncio coroutine"""
    @wraps(func)
    async def inner(*args, **kwargs):
        return await io_bound(func, *args, **kwargs)
    return inner


async def cpu_bound(callback: Callable, *args: any):
    if rosys.is_stopping():
        return
    with cpu():
        try:
            loop = asyncio.get_running_loop()
            return await loop.run_in_executor(process_pool, callback, *args)
        except BrokenProcessPool:
            pass
        except RuntimeError as e:
            if 'cannot schedule new futures after shutdown' not in str(e):
                raise
        except asyncio.exceptions.CancelledError:
            pass


@contextmanager
def cpu() -> Generator[None, None, None]:
    id = str(uuid.uuid4())
    running_cpu_bound_processes.append(id)
    try:
        yield
    finally:
        running_cpu_bound_processes.remove(id)


async def sh(command: list[str] | str, timeout: Optional[float] = 1, shell: bool = False, working_dir: Optional[str] = None) -> str:
    """executes a shell command

    Args:
        command: a sequence of program arguments as subprocess.Popen requires or full string
        shell: whether a subshell should be launched (default is `False`, for speed, use `True` if you need file globbing or other features)

    Returns:
        stdout
    """
    def popen() -> str:
        if shell:  # convert to string
            cmd = ' '.join(command) if isinstance(command, list) else command
        else:  # convert to list
            cmd = shlex.split(command) if isinstance(command, str) else command
        if timeout is not None:
            if shell:
                cmd = f'timeout {timeout} {cmd}'
            else:
                cmd = ['timeout', str(timeout)] + cmd
        # log.info(f'running sh: "{cmd}"')
        proc = subprocess.Popen(
            cmd,
            cwd=working_dir,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.STDOUT,
            shell=shell,
            start_new_session=True,
        )
        running_sh_processes.append(proc)
        stdout, *_ = proc.communicate()
        _kill(proc)
        running_sh_processes.remove(proc)
        return stdout.decode('utf-8')
    if rosys.is_stopping():
        return
    return await io_bound(popen)


def _kill(proc: Popen) -> None:
    try:
        os.killpg(os.getpgid(proc.pid), signal.SIGTERM)
        log.info(f'killed {proc.pid}')
    except ProcessLookupError:
        pass


def tear_down() -> None:
    # stopping process as described in https://www.cloudcity.io/blog/2019/02/27/things-i-wish-they-told-me-about-multiprocessing-in-python/
    log.info('teardown thread_pool')
    thread_pool.shutdown(wait=False, cancel_futures=True)
    [_kill(p) for p in running_sh_processes]
    running_sh_processes.clear()
    if not rosys.is_test:
        log.info('teardown process_pool')
        [p.kill() for p in process_pool._processes.values()]
        process_pool.shutdown(wait=True, cancel_futures=True)
    log.info('teardown complete')



from copy import deepcopy
from dataclasses import dataclass
from typing import Optional

import numpy as np

from .. import rosys
from .calibration import Calibration
from .camera import Camera
from .camera_provider import CameraProvider

ProjectionCoordinates = list[list[Optional[list[float]]]]


@dataclass(slots=True, kw_only=True)
class Projection:
    camera_id: str
    camera_calibration: Calibration
    coordinates: ProjectionCoordinates


class CameraProjector:
    """The camera projector computes a grid of projected image points on the ground plane.

    It is mainly used for visualization purposes.
    """

    def __init__(self, camera_provider: CameraProvider) -> None:
        self.camera_provider = camera_provider

        self.projections: dict[str, Projection] = {}

        rosys.on_repeat(self.step, 1.0)

    async def step(self) -> None:
        for id in list(self.projections):
            if id not in self.camera_provider.cameras:
                del self.projections[id]

        for id, camera in self.camera_provider.cameras.items():
            if not camera.calibration:
                continue
            if id in self.projections and self.projections[id].camera_calibration == camera.calibration:
                continue
            self.projections[id] = Projection(
                camera_id=id,
                camera_calibration=deepcopy(camera.calibration),
                coordinates=self.project(camera),
            )

    @staticmethod
    def project(camera: Camera, rows: int = 12, columns: int = 16) -> ProjectionCoordinates:
        c, r = np.meshgrid(np.linspace(0, camera.calibration.intrinsics.size.width, columns),
                           np.linspace(0, camera.calibration.intrinsics.size.height, rows))
        image_points = np.stack((c.flatten(), r.flatten()), axis=1)
        floor_points = camera.calibration.project_array_from_image(image_points).reshape(rows, columns, 3)
        invalid = np.isnan(floor_points).any(axis=2)
        return [
            [
                None if invalid[r, c] else [point[0], point[1]]
                for c, point in enumerate(row)
            ]
            for r, row in enumerate(floor_points)
        ]

    @staticmethod
    def allclose(array1: ProjectionCoordinates, array2: ProjectionCoordinates) -> bool:
        is_none1 = [point is None for row in array1 for point in row]
        is_none2 = [point is None for row in array2 for point in row]
        if is_none1 != is_none2:
            return False
        values1 = [point for row in array1 for point in row if point is not None]
        values2 = [point for row in array2 for point in row if point is not None]
        return np.allclose(values1, values2)



from __future__ import annotations

import logging
from dataclasses import dataclass, field
from typing import Optional

import cv2
import numpy as np

from ..geometry import Point, Point3d, Rotation
from .image import ImageSize


@dataclass(slots=True, kw_only=True)
class Intrinsics:
    matrix: list[list[float]]
    distortion: list[float]
    rotation: Rotation
    size: ImageSize


log = logging.getLogger('rosys.world.calibration')


@dataclass(slots=True, kw_only=True)
class Extrinsics:
    rotation: Rotation = field(default_factory=lambda: Rotation.from_euler(np.pi, 0, 0))
    translation: list[float] = field(default_factory=lambda: [0.0, 0.0, 1.0])


@dataclass(slots=True, kw_only=True)
class Calibration:
    intrinsics: Intrinsics
    extrinsics: Extrinsics = field(default_factory=lambda: Extrinsics())

    @property
    def rotation(self) -> Rotation:
        return self.extrinsics.rotation * self.intrinsics.rotation

    @property
    def rotation_array(self) -> np.ndarray:
        return np.dot(self.extrinsics.rotation.R, self.intrinsics.rotation.R)

    def project_to_image(self, world_point: Point3d) -> Point:
        world_points = np.array([world_point.tuple], dtype=np.float32)
        R = self.rotation_array
        Rod = cv2.Rodrigues(R.T)[0]
        t = -R.T @ self.extrinsics.translation
        K = np.array(self.intrinsics.matrix)
        D = np.array(self.intrinsics.distortion, dtype=float)
        image_points, _ = cv2.projectPoints(world_points, Rod, t, K, D)
        return Point(x=image_points[0, 0, 0], y=image_points[0, 0, 1])

    def project_array_to_image(self, world_points: np.ndarray) -> np.ndarray:
        R = self.rotation_array
        Rod = cv2.Rodrigues(R.T)[0]
        t = -R.T @ self.extrinsics.translation
        K = np.array(self.intrinsics.matrix)
        D = np.array(self.intrinsics.distortion, dtype=float)
        image_points, _ = cv2.projectPoints(world_points, Rod, t, K, D)
        return image_points.reshape(-1, 2)

    def project_from_image(self, image_point: Point, target_height: float = 0) -> Optional[Point3d]:
        K = np.array(self.intrinsics.matrix)
        D = np.array(self.intrinsics.distortion)
        image_points = np.array(image_point.tuple, dtype=np.float32).reshape(1, 1, 2)
        image_points_ = cv2.undistortPoints(image_points, K, D).reshape(-1, 2)
        image_points__ = cv2.convertPointsToHomogeneous(image_points_).reshape(-1, 3)
        objPoints = image_points__ @ self.rotation_array.T
        Z = self.extrinsics.translation[-1]
        t = np.array(self.extrinsics.translation)
        floor_points = t.T - objPoints * (Z - target_height) / objPoints[:, 2:]

        reprojection = self.project_to_image(Point3d(x=floor_points[0, 0], y=floor_points[0, 1], z=target_height))
        if objPoints[0, -1] * np.sign(Z) > 0 or reprojection.distance(image_point) > 2:
            log.warning(f'reprojection failed with {reprojection.distance(image_point)} px')
            return None

        return Point3d(x=floor_points[0, 0], y=floor_points[0, 1], z=target_height)

    def project_array_from_image(self, image_points: np.ndarray, target_height: float = 0) -> np.ndarray:
        K = np.array(self.intrinsics.matrix)
        D = np.array(self.intrinsics.distortion)
        image_points_ = cv2.undistortPoints(image_points.astype(np.float32), K, D).reshape(-1, 2)
        image_points__ = cv2.convertPointsToHomogeneous(image_points_).reshape(-1, 3)
        objPoints = image_points__ @ self.rotation_array.T
        Z = self.extrinsics.translation[-1]
        t = np.array(self.extrinsics.translation)
        floor_points = t.T - objPoints * (Z - target_height) / objPoints[:, 2:]

        reprojection = self.project_array_to_image(floor_points)
        sign = objPoints[:, -1] * np.sign(Z)
        distance = np.linalg.norm(reprojection - image_points, axis=1)
        floor_points[np.logical_or(sign > 0, distance > 2), :] = np.nan

        return floor_points

    @staticmethod
    def from_points(
            world_points: list[Point3d], image_points: list[Point], image_size: ImageSize, f0: float) -> Calibration:
        world_points = np.array([p.tuple for p in world_points], dtype=np.float32).reshape(1, -1, 3)
        image_points = np.array([p.tuple for p in image_points], dtype=np.float32).reshape(1, -1, 2)
        K0 = np.array([[f0, 0, image_size.width / 2], [0, f0, image_size.height / 2], [0, 0, 1]], dtype=np.float32)
        flags = cv2.CALIB_USE_INTRINSIC_GUESS
        _, K, D, rvecs, tvecs = cv2.calibrateCamera(world_points, image_points, image_size.tuple, K0, None, flags=flags)

        rotation0 = Rotation(R=np.eye(3).tolist())
        intrinsics = Intrinsics(matrix=K.tolist(), distortion=D.tolist()[0], rotation=rotation0, size=image_size)

        rotation = Rotation.from_rvec(rvecs[0]).T
        translation = (-np.array(rotation.R).dot(tvecs)).flatten().tolist()
        extrinsics = Extrinsics(rotation=rotation, translation=translation)

        return Calibration(intrinsics=intrinsics, extrinsics=extrinsics)



import asyncio
import logging
import os
import socket

from nicegui import ui

log = logging.getLogger('rosys.wifi')


def has_internet() -> bool:
    """Returns True if there's a connection"""

    IP_ADDRESS_LIST = [
        '1.1.1.1',  # Cloudflare
        '1.0.0.1',
        '8.8.8.8',  # Google DNS
        '8.8.4.4',
        '208.67.222.222',  # Open DNS
        '208.67.220.220',
    ]

    port = 53
    timeout = 3

    for host in IP_ADDRESS_LIST:
        try:
            socket.setdefaulttimeout(timeout)
            socket.socket(socket.AF_INET, socket.SOCK_STREAM).connect((host, port))
            return True
        except socket.error:
            pass
    else:
        log.warning('No internet connection')
        return False


class WifiButton(ui.button):
    """The WiFi button indicates the current connectivity state and allows setting a new WiFi connection."""

    def __init__(self) -> None:
        with ui.dialog() as self.dialog, ui.card():
            self.status = ui.label()
            self.ssid = ui.input('ssid')
            self.password = ui.input('password')
            with ui.row():
                ui.button('Add', on_click=self.add)
                ui.button('Close', on_click=self.dialog.close)

        super().__init__(on_click=self.dialog.open)
        self.props('icon=network_wifi')
        ui.timer(5, callback=self.update_wifi_status)

    def add(self) -> None:
        log.info(f'adding {self.ssid.value}, {self.password.value}')
        wifi_configs = os.path.expanduser(f'~/.rosys/wifi')
        os.makedirs(wifi_configs, exist_ok=True)
        # NOTE a daemon on the host system must watch the .rosys/wifi dir and reload the configuration with nmcli or similar
        with open(f'{wifi_configs}/{self.ssid.value}', 'w') as f:
            f.write(self.password.value)
        self.dialog.close()

    async def update_wifi_status(self) -> None:
        if await asyncio.get_event_loop().run_in_executor(None, has_internet):
            self.props('color=green')
            self.status.set_text(f'Robot is connected to the internet.')
        else:
            self.props('color=red')
            self.status.set_text(f'Robot is not connected to the internet.')



from .drivable import Drivable
from .driver import Driver
from .driver_object import DriverObject as driver_object
from .joystick_ import Joystick as joystick
from .keyboard_control_ import KeyboardControl as keyboard_control
from .odometer import Odometer, VelocityProvider
from .path_segment import PathSegment
from .robot_object_ import RobotObject as robot_object
from .steerer import Steerer



import logging
from typing import TYPE_CHECKING, Optional

import cv2
import numpy as np
from fastapi import Response
from nicegui import app

from .. import run
from .camera import Camera
from .image import Image

if TYPE_CHECKING:
    from .camera_provider import CameraProvider

log = logging.getLogger('rosys.image_route')


def create_image_route(camera_provider: 'CameraProvider') -> None:
    async def get_image(camera_id: str, timestamp: str, shrink: int = 1) -> Response:
        return await _get_image(camera_provider.cameras, camera_id, timestamp, shrink)
    app.add_api_route('/' + camera_provider.base_path + '/{camera_id}/{timestamp}', get_image)
    app.add_api_route('/' + camera_provider.base_path + '/placeholder', _get_placeholder)


def _get_placeholder() -> Response:
    return Response(content=Image.create_placeholder('no image').data, media_type='image/jpeg')


async def _get_image(cameras: dict[str, Camera], camera_id: str, timestamp: str, shrink: int = 1) -> Response:
    try:
        camera = cameras.get(camera_id)
        if not camera:
            return Response(content='Camera not found', status_code=404)
        jpeg = await _try_get_jpeg(camera, timestamp, shrink)
        if not jpeg:
            return Response(content='Image not found', status_code=404)
        return Response(content=jpeg, headers={'cache-control': 'max-age=7776000'}, media_type='image/jpeg')
    except Exception:
        log.exception('could not get image')
        raise


async def _try_get_jpeg(camera: Camera, timestamp: str, shrink: int) -> Optional[bytes]:
    for image in reversed(camera.images):
        if str(image.time) == timestamp and image.data is not None:
            jpeg = image.data
            if shrink != 1:
                array = np.frombuffer(image.data, dtype=np.uint8)
                if array is None:
                    return
                jpeg = await run.cpu_bound(_shrink, shrink, array)
            return jpeg


def _shrink(factor: int, array: np.ndarray) -> bytes:
    decoded = cv2.imdecode(array, cv2.IMREAD_COLOR)
    if decoded is None:
        return
    img = decoded[::factor, ::factor]
    return cv2.imencode('.jpg', img, [int(cv2.IMWRITE_JPEG_QUALITY), 60])[1].tobytes()



import time
from typing import Optional

import numpy as np
import scipy.interpolate

from ..geometry import Point
from .obstacle_map import ObstacleMap


class DistanceMap:

    def __init__(self, obstacle_map: ObstacleMap, target: Point, deadline: Optional[float] = None):
        self.grid = obstacle_map.grid
        scaled_obstacle_map = obstacle_map.map

        self.INF = 1000000
        self.F = 100
        dx = self.F * self.grid.bbox[2] / self.grid.size[1]
        dy = self.F * self.grid.bbox[3] / self.grid.size[0]
        self.DX = int(dx)
        self.DY = int(dy)
        self.DD = int(np.sqrt(dx**2 + dy**2))

        scaled_inf = self.F * self.INF

        d = np.zeros(self.grid.size[:2], dtype=int)
        d.fill(scaled_inf)
        row, col = self.grid.to_grid(target.x, target.y)
        for r in [np.floor(row), np.ceil(row)]:
            for c in [np.floor(col), np.ceil(col)]:
                d[int(r), int(c)] = np.sqrt((row - r)**2 + (col - c)**2)

        old_sum = d.sum()
        while True:
            d_dx = d + self.DX
            d_dy = d + self.DY
            d_dd = d + self.DD
            d[1:-1, 1:-1] = np.minimum(
                d[1:-1, 1:-1],
                np.minimum(
                    np.minimum(
                        np.minimum(d_dy[:-2, 1:-1], d_dy[+2:, 1:-1]),
                        np.minimum(d_dx[1:-1, :-2], d_dx[1:-1, +2:])
                    ),
                    np.minimum(
                        np.minimum(d_dd[+2:, +2:], d_dd[+2:, :-2]),
                        np.minimum(d_dd[:-2, +2:], d_dd[:-2, :-2])
                    )
                )
            )
            d[scaled_obstacle_map] = scaled_inf
            new_sum = d.sum()
            if new_sum == old_sum:
                break
            old_sum = new_sum
            if deadline and time.time() > deadline:
                raise TimeoutError('distance map creation took too long')

        self.map = d.astype(float) / self.F
        gy, gx = np.gradient(self.map)

        rows = np.arange(self.grid.size[0])
        cols = np.arange(self.grid.size[1])
        xx, yy = self.grid.from_grid(rows, cols)
        self._interp = scipy.interpolate.interp2d(xx, yy, self.map, kind='linear')
        self._grad_x = scipy.interpolate.interp2d(xx, yy, gx, kind='linear')
        self._grad_y = scipy.interpolate.interp2d(xx, yy, gy, kind='linear')

        self.map[self.map >= self.INF] = np.inf

    def interpolate(self, x, y):
        result = self._interp(x, y)
        result[result >= self.INF / self.F] = np.inf
        return result

    def gradient(self, x, y):
        result_x = self._grad_x(x, y)
        result_y = self._grad_y(x, y)
        result_x[result_x <= -self.INF / self.F] = -np.inf
        result_y[result_y <= -self.INF / self.F] = -np.inf
        result_x[result_x >= self.INF / self.F] = np.inf
        result_y[result_y >= self.INF / self.F] = np.inf
        return result_x, result_y



import asyncio
from datetime import datetime, timedelta
from typing import Optional

import socketio
import socketio.exceptions

from .. import background_tasks, persistence, rosys
from .detections import BoxDetection, Detections, PointDetection, SegmentationDetection
from .detector import Autoupload, Detector
from .image import Image
from .uploads import Uploads


class DetectorHardware(Detector):
    """This detector communicates with a [YOLO detector](https://hub.docker.com/r/zauberzeug/yolov5-detector) via Socket.IO.

    It automatically connects and reconnects, submits and receives detections and sends images that should be uploaded to the [Zauberzeug Learning Loop](https://zauberzeug.com/learning-loop.html).
    """

    def __init__(self, *, port: int = 8004) -> None:
        super().__init__()

        self.sio = socketio.AsyncClient()
        self.is_detecting: bool = False
        self.next_image: Optional[Image] = None
        self.port = port
        self.timeout_count = 0
        self._uploads = Uploads()

        @self.sio.on('disconnect')
        def on_sio_disconnect() -> None:
            self.log.warning(f'sio disconnect on port {port}')

        @self.sio.on('connect_error')
        def on_sio_connect_error(err) -> None:
            self.log.warning(f'sio connect error on {port}: {err}')

        rosys.on_repeat(self.step, 1.0)

    @property
    def uploads(self) -> Uploads:
        return self._uploads

    @property
    def is_connected(self) -> bool:
        return self.sio.connected

    async def step(self) -> None:
        if not self.is_connected:
            self.log.info(f'trying reconnect {self.port}')
            if not await self.connect():
                self.log.exception(f'connection to {self.port} failed; trying again')
                await rosys.sleep(3.0)
                return

        await self.try_start_one_upload()
        await self.upload_priority_queue()

    async def connect(self) -> bool:
        try:
            url = f'ws://localhost:{self.port}'
            self.log.info(f'connecting to detector at {url}')
            await self.sio.connect(url, socketio_path='/ws/socket.io', wait_timeout=3.0)
            self.log.info('connected successfully')
            return True
        except Exception:
            return False

    async def disconnect(self) -> None:
        await self.sio.disconnect()

    async def try_start_one_upload(self) -> None:
        if datetime.now() < self.uploads.last_upload + timedelta(minutes=self.uploads.minimal_minutes_between_uploads):
            return

        upload_images = self.uploads.get_queued()
        if upload_images:
            background_tasks.create(self.upload(upload_images[0]), name='upload_image')
            self.uploads.queue.clear()  # old images should not be uploaded later when the robot is inactive
            self.uploads.last_upload = datetime.now()

    async def upload_priority_queue(self) -> None:
        upload_images = self.uploads.get_priority_queued()
        if upload_images:
            async def upload_priority_images():
                for image in upload_images:
                    await self.upload(image)
            background_tasks.create(upload_priority_images(), name='upload_priority_images')
            self.uploads.priority_queue.clear()

    async def upload(self, image: Image) -> None:
        await super().upload(image)
        try:
            self.log.info(f'uploading to port {self.port}')
            await self.sio.emit('upload', {'image': image.data, 'mac': image.camera_id})
        except Exception:
            self.log.exception(f'could not upload {image.id}')

    async def detect(self, image: Image, autoupload: Autoupload = Autoupload.FILTERED, tags: list[str] = []) -> None:
        """Runs detections on the image. Afterwards the `image.detections` property is filled."""
        if not self.is_connected:
            return

        self.next_image = image
        if self.is_detecting:
            return

        while self.next_image is not None and not rosys.is_stopping():
            current_image = self.next_image
            self.next_image = None
            if current_image.is_broken:
                continue
            try:
                self.is_detecting = True
                result: dict = await self.sio.call('detect', {
                    'image': current_image.data,
                    'mac': current_image.camera_id,
                    'autoupload': autoupload.value,
                    'tags': tags,
                }, timeout=3)
                if current_image.is_broken:  # NOTE: image can be marked broken while detection is underway
                    continue
                current_image.detections = Detections(
                    boxes=[persistence.from_dict(BoxDetection, d) for d in result.get('box_detections', [])],
                    points=[persistence.from_dict(PointDetection, d) for d in result.get('point_detections', [])],
                    segmentations=[
                        persistence.from_dict(SegmentationDetection, d)
                        for d in result.get('segmentation_detections', [])
                    ],
                )
            except socketio.exceptions.TimeoutError:
                self.log.debug(f'detection for {current_image.id} on {self.port} took too long')
                self.timeout_count += 1
            except asyncio.exceptions.CancelledError:
                self.log.debug(f'task has been cancelled')
            except Exception:
                self.log.exception(f'could not detect {current_image.id}')
            else:
                self.timeout_count = 0
                if not current_image.is_broken:
                    self.NEW_DETECTIONS.emit(current_image)
            finally:
                self.is_detecting = False
                if self.timeout_count > 5:
                    await self.disconnect()

    def __str__(self) -> str:
        return f'{type(self).__name__} ({"connected" if self.is_connected else "disconnected"})'



from __future__ import annotations

from dataclasses import dataclass, field

from ..geometry import Point


@dataclass(slots=True, kw_only=True)
class Detection:
    category_name: str
    model_name: str
    confidence: float
    x: float
    y: float
    uuid: str = ''  # NOTE this is used to keep track of simulated objects

    @property
    def cx(self) -> float:
        raise NotImplementedError()

    @property
    def cy(self) -> float:
        raise NotImplementedError()

    @property
    def center(self) -> Point:
        return Point(x=self.cx, y=self.cy)


@dataclass(slots=True, kw_only=True)
class BoxDetection(Detection):
    width: float
    height: float

    @property
    def cx(self) -> float:
        return self.x + self.width / 2

    @property
    def cy(self) -> float:
        return self.y + self.height / 2

    def __str__(self) -> str:
        return f'x:{self.x:.0f} y:{self.y:.0f}, w:{self.width:.0f} h:{self.height:.0f} cat:{self.category_name} conf:{self.confidence:.2f}'

    def to_svg(self, shrink: int = 1) -> str:
        color = 'red'
        x = self.x / shrink
        y = self.y / shrink
        return f'<rect x="{x}" y="{y}" width="{self.width/shrink}" height="{self.height/shrink}" stroke-width="2" stroke="{color}" fill="none" />' \
            f'<text x="{x}" y="{y - 7}" text-anchor="start" stroke="{color}" fill="{color}" font-size="10">{self.category_name} ({int(self.confidence*100)}%)</text>'


@dataclass(slots=True, kw_only=True)
class Shape:
    points: list[Point]

    @staticmethod
    def from_str(points: str) -> Shape:
        p = points.split(',')
        points = [Point(int(p[i]), int(p[i+1])) for i in range(0, len(p), 2)]
        return Shape(points=points)

    def __str__(self) -> str:
        return ', '. join([f'x:{p.x:.0f} y:{p.y:.0f}' for p in self.points])


@dataclass(slots=True, kw_only=True)
class SegmentationDetection:
    category_name: str
    model_name: str
    confidence: float
    shape: Shape

    @staticmethod
    def from_dict(detection: dict) -> SegmentationDetection:
        return SegmentationDetection(
            detection['category_name'],
            detection['model_name'],
            detection['confidence'],
            Shape.from_str(detection['shape']))

    def __str__(self) -> str:
        return f'shape:{self.shape.__str__} cat:{self.category_name} conf:{self.confidence:.2f}'

    def to_svg(self, shrink: int = 1) -> str:
        color = 'green'
        d = ' '.join([f"{'L' if i > 0 else 'M'} {p.x/shrink} {p.y/shrink}" for i, p in enumerate(self.shape.points)])
        d += ' Z'
        p0 = self.shape.points[0]
        return f'<path d="{d}" stroke-width="2" stroke="{color}" fill="none" />' \
            f'<text x="{p0.x / shrink + 10}" y="{p0.y / shrink + 4}" text-anchor="start" stroke="{color}" fill="{color}" font-size="12" font-weight="light">{self.category_name} ({int(self.confidence*100)}%)</text>'


@dataclass(slots=True, kw_only=True)
class PointDetection(Detection):

    @property
    def cx(self) -> float:
        return self.x

    @property
    def cy(self) -> float:
        return self.y

    def __str__(self) -> str:
        return f'x:{self.x:.0f} y:{self.y:.0f} cat:{self.category_name} conf:{self.confidence:.2f}'

    def to_svg(self, shrink: int = 1) -> str:
        color = 'red'
        x = self.x / shrink
        y = self.y / shrink
        return f'<circle cx="{x}" cy="{y}" r="4" stroke-width="2" stroke="{color}" fill="none" />' \
            f'<text x="{x + 10}" y="{y + 4}" text-anchor="start" stroke="{color}" fill="{color}" font-size="12" font-weight="light">{self.category_name} ({int(self.confidence*100)}%)</text>'


@dataclass(slots=True, kw_only=True)
class Detections:
    boxes: list[BoxDetection] = field(default_factory=list)
    points: list[PointDetection] = field(default_factory=list)
    segmentations: list[SegmentationDetection] = field(default_factory=list)

    def to_svg(self, shrink: int = 1) -> str:
        return '\n'.join(
            b.to_svg(shrink) for b in self.boxes) + '\n' + '\n'.join(
            p.to_svg(shrink) for p in self.points) + '\n' + '\n'.join(
            s.to_svg(shrink) for s in self.segmentations)



from .module import ModuleHardware
from .robot_brain import RobotBrain


class SerialHardware(ModuleHardware):
    """The serial module represents a serial connection with another device."""

    def __init__(self, robot_brain: RobotBrain, *,
                 name: str = 'serial',
                 rx_pin: int = 26,
                 tx_pin: int = 27,
                 baud: int = 115_200,
                 num: int = 1) -> None:

        self.name = name
        lizard_code = f'{name} = Serial({rx_pin}, {tx_pin}, {baud}, {num})'
        super().__init__(robot_brain=robot_brain, lizard_code=lizard_code)



import logging
import re
import shutil
from dataclasses import dataclass

from ... import rosys


@dataclass(slots=True, kw_only=True)
class NetworkStats:
    tx_errors: int
    tx_dropped: int
    rx_errors: int
    rx_dropped: int

    def __gt__(self, other) -> bool:
        return self.tx_errors > other.tx_errors or \
            self.tx_dropped > other.tx_dropped or \
            self.rx_errors > other.rx_errors or \
            self.rx_dropped > other.rx_dropped

    def msg(self) -> str:
        return f'{self.tx_errors}/{self.rx_errors} errors and {self.tx_dropped}/{self.rx_dropped} dropped'


class NetworkMonitor:

    def __init__(self) -> None:
        self.log = logging.getLogger('rosys.network_monitor')
        self.interfaces: dict[str, NetworkStats] = {}

        rosys.on_repeat(self.step, 60)

    @staticmethod
    def is_operable() -> bool:
        return shutil.which('ip') is not None

    async def step(self) -> None:
        output = await rosys.run.sh(['ip', '-s', 'a'])
        for interface in NetworkMonitor.split_interfaces(output):
            name = interface[:interface.index(':')]
            lines = interface.split('\n')
            stats = NetworkStats(
                tx_errors=int(lines[-1].split()[2]),
                tx_dropped=int(lines[-1].split()[3]),
                rx_errors=int(lines[-3].split()[2]),
                rx_dropped=int(lines[-3].split()[3]),
            )
            if name not in self.interfaces:
                self.interfaces[name] = stats
            else:
                if stats > self.interfaces[name]:
                    msg = name + ' has ' + stats.msg()
                    self.log.warning(msg)
                    rosys.notify(msg, 'warning')
                    self.interfaces[name] = stats

    @staticmethod
    def split_interfaces(output: str) -> list[str]:
        spacing = re.sub('^[0-9]+: ', '\n', output, flags=re.MULTILINE)
        return spacing.strip().split('\n\n')



import asyncio
import gc
from collections import Counter
from pathlib import Path
from typing import Optional

import objgraph
from nicegui import ui

import rosys

SVG_FILE = Path('/tmp/rosys_objgraph.svg')


def objgraph_page() -> None:

    @ui.page('/objgraph')
    def page() -> None:
        async def update_graph() -> None:
            class_search_result.content = await create_graph(class_search.value)

        async def refresh_stats() -> None:
            loop = asyncio.get_running_loop()
            growth, leaking, overall, counts = await loop.run_in_executor(None, get_objgraph_stats)
            await update_graph()
            growth_label.set_text(growth)
            leaking_label.set_text(leaking)
            overall_label.set_text(overall)
            highest = [f'{c[1]}: {c[0]}' for c in counts.most_common()[:5]]
            counts_markdown.set_content('new obj occurrence: ' + '\n'.join(highest))
            most_common_search_result.content = SVG_FILE.read_text() if counts else ''

        t = ui.timer(10, refresh_stats, active=False)
        ui.switch('track objects (every 10 s)').bind_value_to(t, 'active')
        with ui.column():
            growth_label = ui.label()
            leaking_label = ui.label()
            overall_label = ui.label()
            counts_markdown = ui.markdown()

        class_search = ui.input('Search Class', value='Client')
        ui.button('Update Graph', on_click=update_graph)
        class_search_result = ui.html()
        most_common_search_result = ui.html()


@rosys.run.awaitable
def create_graph(search_term: str) -> str:
    gc.collect()
    objects = objgraph.by_type(search_term)
    content = ''
    for obj in objects:
        chain = objgraph.find_backref_chain(obj, objgraph.is_proper_module)
        objgraph.show_chain(chain, filename=str(SVG_FILE))
        content += SVG_FILE.read_text()
    return content


def get_objgraph_stats() -> tuple[str, str, str, Optional[Counter[str]]]:
    gc.collect()
    growth_objs = objgraph.growth(4)
    growth = 'object growth every 10 sec: '
    for obj in growth_objs:
        growth += f'{obj[0]}: +{obj[2]}  '
    leaking = 'obj without refs: '
    for o, c in list(objgraph.typestats(objgraph.get_leaking_objects()).items())[:4]:
        leaking += f'{o}: {c}  '
    overall = 'overall obj count: '
    for o, c in list(objgraph.typestats().items())[:4]:
        overall += f'{o}: {c}  '
    if growth_objs:
        biggest_growth = growth_objs[0][0]
        new = objgraph.at_addrs(objgraph.get_new_ids()[biggest_growth])
        counts = Counter(str(i) for i in new)
        if counts:
            most_common = counts.most_common()[0][0]
            unique = {str(i): i for i in new}
            chain = objgraph.find_backref_chain(unique[most_common], objgraph.is_proper_module)
            objgraph.show_chain(chain, filename=str(SVG_FILE))
    else:
        counts = None

    return growth, leaking, overall, counts



import io
import logging
import re
import shutil
from dataclasses import dataclass, field
from typing import Any, Optional

import cv2
import numpy as np
import PIL

from .. import persistence, rosys
from ..geometry import Rectangle
from .camera_provider import CameraProvider
from .image import Image, ImageSize
from .usb_camera import ImageRotation, UsbCamera

MJPG = cv2.VideoWriter_fourcc(*'MJPG')
SCAN_INTERVAL = 10


@dataclass(slots=True, kw_only=True)
class Device:
    uid: str
    video_id: int
    capture: Optional[Any] = None  # cv2.VideoCapture device
    exposure_min: int = 0
    exposure_max: int = 0
    exposure_default: int = 0
    has_manual_exposure: bool = False
    last_state: dict = field(default_factory=dict)
    video_formats: set[str] = field(default_factory=set)


def process_jpeg_image(data: bytes, rotation: ImageRotation, crop: Rectangle = None) -> bytes:
    image = PIL.Image.open(io.BytesIO(data))
    if crop is not None:
        image = image.crop((int(crop.x), int(crop.y), int(crop.x+crop.width), int(crop.y+crop.height)))
    if rotation == ImageRotation.LEFT:
        image = cv2.rotate(image, cv2.ROTATE_90_COUNTERCLOCKWISE)
    elif rotation == ImageRotation.RIGHT:
        image = cv2.rotate(image, cv2.ROTATE_90_CLOCKWISE)
    elif rotation == ImageRotation.UPSIDE_DOWN:
        image = cv2.rotate(image, cv2.ROTATE_180)
    img_byte_arr = io.BytesIO()
    image.save(img_byte_arr, format='JPEG')
    return img_byte_arr.getvalue()


def process_ndarray_image(image: np.ndarray, rotation: ImageRotation, crop: Rectangle = None) -> bytes:
    if crop is not None:
        image = image[int(crop.y):int(crop.y+crop.height), int(crop.x):int(crop.x+crop.width)]
    if rotation == ImageRotation.LEFT:
        image = cv2.rotate(image, cv2.ROTATE_90_COUNTERCLOCKWISE)
    elif rotation == ImageRotation.RIGHT:
        image = cv2.rotate(image, cv2.ROTATE_90_CLOCKWISE)
    elif rotation == ImageRotation.UPSIDE_DOWN:
        image = cv2.rotate(image, cv2.ROTATE_180)
    return cv2.imencode('.jpg', image)[1].tobytes()


def to_bytes(image: Any) -> bytes:
    return image[0].tobytes()


class UsbCameraProviderHardware(CameraProvider):
    """This module collects and provides real USB cameras.

    Camera devices are discovered through video4linux (v4l) and accessed with openCV.
    Therefore the program v4l2ctl and openCV (including python bindings) must be available.
    """

    def __init__(self) -> None:
        super().__init__()

        self.log = logging.getLogger('rosys.usb_camera_provider')

        self.devices: dict[str, Device] = {}
        self.last_scan: Optional[float] = None
        self._cameras: dict[str, UsbCamera] = {}

        rosys.on_shutdown(self.shutdown)
        rosys.on_repeat(self.capture_images, 0.2)
        rosys.on_repeat(self.update_parameters, 1)
        rosys.on_repeat(self.update_device_list, 1)
        rosys.on_repeat(lambda: self.prune_images(max_age_seconds=1.0), 5.0)

        self.needs_backup: bool = False
        persistence.register(self)

    @property
    def cameras(self) -> dict[str, UsbCamera]:
        return self._cameras

    def backup(self) -> dict:
        return {'cameras': persistence.to_dict(self._cameras)}

    def restore(self, data: dict[str, Any]) -> None:
        persistence.replace_dict(self._cameras, UsbCamera, data.get('cameras', {}))

    async def capture_images(self) -> None:
        for uid, camera in self._cameras.items():
            if not camera.active:
                if uid in self.devices and self.devices[uid].capture is not None:
                    await self.deactivate(camera)
                continue
            try:
                if uid not in self.devices:
                    continue
                device = self.devices[uid]
                if device.capture is None:
                    await self.activate(uid)
                if device.capture is None:
                    self.log.warning(f'unexpected missing capture handle for {uid}')
                    continue
                image = await rosys.run.io_bound(self.capture_image, uid)
                if image is None:
                    await self.deactivate(camera)
                    continue
                if 'MJPG' in device.video_formats:
                    bytes = await rosys.run.io_bound(to_bytes, image)
                    if camera.crop or camera.rotation != ImageRotation.NONE:
                        bytes = await rosys.run.cpu_bound(process_jpeg_image, bytes, camera.rotation, camera.crop)
                else:
                    bytes = await rosys.run.cpu_bound(process_ndarray_image, image, camera.rotation, camera.crop)
                size = camera.resolution or ImageSize(width=800, height=600)
                camera.images.append(Image(camera_id=uid, data=bytes, time=rosys.time(), size=size))
            except Exception:
                self.log.exception(f'could not capture image from {uid}')
                await self.deactivate(camera)

    async def update_parameters(self) -> None:
        for uid, camera in self._cameras.items():
            if camera.active and uid in self.devices:
                await rosys.run.io_bound(self.set_parameters, camera, self.devices[uid])

    def capture_image(self, id) -> Any:
        _, image = self.devices[id].capture.read()
        return image

    async def activate(self, uid: str) -> None:
        camera = self._cameras[uid]
        device = self.devices[uid]
        capture = await rosys.run.io_bound(self.get_capture_device, device.video_id)
        if capture is None:
            return
        self.devices[uid].capture = capture
        await rosys.run.io_bound(self.set_parameters, camera, device)
        self.log.info(f'activated {uid}')
        await self.CAMERA_ADDED.call(camera)

    async def deactivate(self, camera: UsbCamera) -> None:
        if camera.id not in self.devices:
            return
        if self.devices[camera.id].capture is not None:
            self.log.info(f'deactivated {camera.id}')
            await rosys.run.io_bound(self.devices[camera.id].capture.release)
        del self.devices[camera.id]

    async def update_device_list(self) -> None:
        if self.last_scan is not None and rosys.time() < self.last_scan + SCAN_INTERVAL:
            return
        self.last_scan = rosys.time()
        output = await rosys.run.sh(['v4l2-ctl', '--list-devices'])
        if output is None:
            return
        output = '\n'.join([s for s in output.split('\n') if not s.startswith('Cannot open device')])
        for infos in output.split('\n\n'):
            match = re.search(r'\((.*)\)', infos)
            if match is None:
                continue
            uid = match.group(1)
            if uid not in self._cameras:
                self._cameras[uid] = UsbCamera(id=uid)
                self.log.info(f'adding camera {uid}')
                await self.CAMERA_ADDED.call(self._cameras[uid])
            lines = infos.splitlines()
            if 'dev/video' not in lines[1]:
                continue
            num = int(lines[1].strip().lstrip('/dev/video'))
            if uid not in self.devices:
                self.devices[uid] = Device(uid=uid, video_id=num)
                await self.load_value_ranges(self.devices[uid])

    def get_capture_device(self, index: int) -> None:
        try:
            capture = cv2.VideoCapture(index)
            if capture is None:
                self.log.error(f'video device {index} is unavailable')
            elif not capture.isOpened():
                self.log.error(f'video device {index} can not be opened')
                capture.release()
            else:
                return capture
        except Exception:
            self.log.exception(f'{index} device failed')

    async def shutdown(self) -> None:
        for camera in self._cameras.values():
            await self.deactivate(camera)
        self.devices.clear()

    @staticmethod
    def is_operable() -> bool:
        return shutil.which('v4l2-ctl') is not None

    def set_parameters(self, camera: UsbCamera, device: Device) -> None:
        camera.fps = int(device.capture.get(cv2.CAP_PROP_FPS))
        if not camera.auto_exposure and camera.exposure is None and device.exposure_max > 0:
            camera.exposure = device.exposure_default / device.exposure_max
        if 'MJPG' in device.video_formats:
            # NOTE enforcing motion jpeg for now
            if device.capture.get(cv2.CAP_PROP_FOURCC) != MJPG:
                device.capture.set(cv2.CAP_PROP_FOURCC, MJPG)
            # NOTE disable video decoding (see https://stackoverflow.com/questions/62664621/read-jpeg-frame-from-mjpeg-camera-without-decoding-in-python-opencv/70869738?noredirect=1#comment110818859_62664621)
            if device.capture.get(cv2.CAP_PROP_CONVERT_RGB) != 0:
                device.capture.set(cv2.CAP_PROP_CONVERT_RGB, 0)
            # NOTE make sure there is no lag (see https://stackoverflow.com/a/30032945/364388)
            if device.capture.get(cv2.CAP_PROP_BUFFERSIZE) != 1:
                device.capture.set(cv2.CAP_PROP_BUFFERSIZE, 1)
        resolution = ImageSize(
            width=int(device.capture.get(cv2.CAP_PROP_FRAME_WIDTH)),
            height=int(device.capture.get(cv2.CAP_PROP_FRAME_HEIGHT))
        )
        if camera.resolution and camera.resolution != resolution:
            self.log.info(f'updating resolution of {camera.id} from {resolution} to {camera.resolution}')
            device.capture.set(cv2.CAP_PROP_FRAME_WIDTH, camera.resolution.width)
            device.capture.set(cv2.CAP_PROP_FRAME_HEIGHT, camera.resolution.height)
        if device.has_manual_exposure:
            auto_exposure = device.capture.get(cv2.CAP_PROP_AUTO_EXPOSURE) == 3
            if camera.auto_exposure and not auto_exposure:
                self.log.info(f'activating auto-exposure for {camera.id}')
                device.capture.set(cv2.CAP_PROP_AUTO_EXPOSURE, 3)  # `v4l2-ctl -L` says "3: Aperture Priority Mode"
            if auto_exposure and not camera.auto_exposure:
                self.log.info(f'deactivating auto-exposure of {camera.id}')
                device.capture.set(cv2.CAP_PROP_AUTO_EXPOSURE, 1)  # `v4l2-ctl -L` says "1: Manual Mode"
            if not camera.auto_exposure and device.exposure_max > 0:
                exposure = device.capture.get(cv2.CAP_PROP_EXPOSURE) / device.exposure_max
                if camera.exposure is not None and camera.exposure != exposure:
                    self.log.info(f'updating exposure of {camera.id} from {exposure} to {camera.exposure})')
                    device.capture.set(cv2.CAP_PROP_EXPOSURE, int(camera.exposure * device.exposure_max))

    async def load_value_ranges(self, device: Device) -> None:
        output = await self.run_v4l(device, '--all')
        if output is None:
            return
        match = re.search(r'exposure_absolute.*: min=(\d*).*max=(\d*).*default=(\d*).*', output)
        if match is not None:
            device.has_manual_exposure = True
            device.exposure_min = int(match.group(1))
            device.exposure_max = int(match.group(2))
            device.exposure_default = int(match.group(3))
        else:
            device.has_manual_exposure = False
        output = await self.run_v4l(device, '--list-formats')
        match = re.finditer(r"$.*'(.*)'.*", output)
        for m in match:
            device.video_formats.add(m.group(1))

    async def run_v4l(self, device: Device, *args) -> None:
        cmd = ['v4l2-ctl', '-d', str(device.video_id)]
        cmd.extend(args)
        return await rosys.run.sh(cmd)



#!/usr/bin/env python3
from __future__ import annotations

import importlib
import logging
import sys
import time

import icecream
import numpy as np
import pylab as pl
from nicegui import ui

from rosys.geometry import Point, Pose
from rosys.pathplanning import plot_tools as pt
from rosys.pathplanning.area import Area
from rosys.pathplanning.delaunay_planner import DelaunayPlanner
from rosys.pathplanning.obstacle import Obstacle
from rosys.pathplanning.planner_process import PlannerSearchCommand
from rosys.pathplanning.robot_renderer import RobotRenderer

icecream.install()

seed = np.random.randint(0, 1000)
seed = 0
np.random.seed(seed)

cmd = PlannerSearchCommand(
    areas=[Area(id='main', outline=[Point(x=-5, y=-5), Point(x=25, y=-5), Point(x=25, y=25), Point(x=-5, y=25)])],
    obstacles=[
        Obstacle(id='0', outline=[Point(x=5, y=-5), Point(x=8, y=-5), Point(x=8, y=15), Point(x=5, y=15)]),
        Obstacle(id='1', outline=[Point(x=12, y=5), Point(x=15, y=5), Point(x=15, y=25), Point(x=12, y=25)]),
    ],
    start=Pose(x=np.random.uniform(-4, 4), y=np.random.uniform(-4, 4), yaw=np.random.uniform(-np.pi, np.pi)),
    goal=Pose(x=np.random.uniform(16, 24), y=np.random.uniform(16, 24), yaw=np.random.uniform(-np.pi, np.pi)),
    deadline=np.inf,
)
robot_outline = [(-0.5, -0.5), (0.5, -0.5), (0.75, 0), (0.5, 0.5), (-0.5, 0.5)]

demo = sys.argv[1] if len(sys.argv) > 1 else ''
if demo:
    i = importlib.import_module(demo.rstrip('.py').replace('/', '.'))
    cmd = i.cmd
    robot_outline = i.robot_outline

planner = DelaunayPlanner(robot_outline)


def run() -> None:
    t = time.time()
    planner.update_map(cmd.areas, cmd.obstacles, [cmd.start, cmd.goal], deadline=time.time()+10.0)
    dt0 = time.time() - t

    t = time.time()
    try:
        path = planner.search(cmd.start, cmd.goal)
    except RuntimeError:
        logging.exception('could not find path')
        path = []
    dt1 = time.time() - t

    with plot:
        pl.clf()
        pl.title(f'map: {dt0:.3f} s, path: {dt1:.3f} s ({seed=})')
        pt.show_obstacle_map(planner.obstacle_map)
        pl.gca().invert_yaxis()
        pl.autoscale(False)
        pl.triplot(planner.tri_points[:, 0], planner.tri_points[:, 1], planner.tri_mesh.simplices, lw=0.1)

        pt.plot_path(path, 'C1')
        robot_renderer = RobotRenderer(robot_outline)
        pt.plot_robot(robot_renderer, (cmd.start.x, cmd.start.y, cmd.start.yaw), 'C0', lw=2)
        pt.plot_robot(robot_renderer, (cmd.goal.x, cmd.goal.y, cmd.goal.yaw), 'C0', lw=2)
        for step in path:
            for t in [0, 1]:
                yaw = step.spline.yaw(t) + np.pi if step.backward else step.spline.yaw(t)
                pt.plot_robot(robot_renderer, (step.spline.x(t), step.spline.y(t), yaw), 'C2', lw=1)


with ui.row():
    plot = ui.plot(figsize=(8, 8))
    with ui.column():
        ui.button('Re-run', on_click=run).props('icon=replay outline')
        ui.label(demo)
run()

ui.run()



import io
import random
import time
from typing import Any, Optional

import numpy as np
import PIL as pil

from rosys import persistence

from .. import rosys
from .camera_provider import CameraProvider
from .image import Image, ImageSize
from .usb_camera import UsbCamera


class UsbCameraProviderSimulation(CameraProvider):
    """This module collects and simulates USB cameras and generates synthetic images.

    In the current implementation the images only contain the camera ID and the current time.
    """

    def __init__(self) -> None:
        super().__init__()

        self._cameras: dict[str, UsbCamera] = {}

        rosys.on_repeat(self.step, 1.0)
        rosys.on_repeat(lambda: self.prune_images(max_age_seconds=1.0), 5.0)

        self.needs_backup: bool = False
        persistence.register(self)

    @property
    def cameras(self) -> dict[str, UsbCamera]:
        return self._cameras

    def backup(self) -> dict:
        return {'cameras': persistence.to_dict(self._cameras)}

    def restore(self, data: dict[str, Any]) -> None:
        persistence.replace_dict(self._cameras, UsbCamera, data.get('cameras', {}))

    async def step(self) -> None:
        for camera in self._cameras.values():
            if not camera.active:
                continue
            assert camera.image_resolution is not None, 'simulated USB cameras should have an image resolution'
            image = Image(time=rosys.time(), camera_id=camera.id, size=camera.image_resolution)
            if rosys.is_test:
                image.data = b'test data'
            else:
                image.data = await rosys.run.cpu_bound(self.create_image_data, camera)
            camera.images.append(image)

    @staticmethod
    def create(uid: str, width: int = 800, height: int = 600, color: Optional[str] = None) -> UsbCamera:
        color = color or f'#{random.randint(0, 0xffffff):06x}'
        return UsbCamera(id=uid, resolution=ImageSize(width=width, height=height), color=color)

    @staticmethod
    def create_calibrated(uid: str, *,
                          width: int = 800, height: int = 600, color: Optional[str] = None,
                          x: float = 0.0, y: float = 0.0, z: float = 1.0,
                          roll: float = np.pi, pitch: float = 0.0, yaw: float = 0.0) -> UsbCamera:
        camera = UsbCameraProviderSimulation.create(uid, width, height, color)
        camera.set_perfect_calibration(width=width, height=height, x=x, y=y, z=z, roll=roll, pitch=pitch, yaw=yaw)
        return camera

    @staticmethod
    def create_image_data(camera: UsbCamera) -> bytes:
        size = camera.image_resolution
        img = pil.Image.new('RGB', size=(size.width, size.height), color=camera.color)
        d = pil.ImageDraw.Draw(img)
        text = f'{camera.id}: {time.time()}'
        d.text((img.width / 2 - len(text) * 3 + 0, img.height / 2 - 5), text, fill=(0, 0, 0))
        d.text((img.width / 2 - len(text) * 3 + 1, img.height / 2 - 4), text, fill=(255, 255, 255))
        img_byte_arr = io.BytesIO()
        img.save(img_byte_arr, format='JPEG')
        return img_byte_arr.getvalue()



from dataclasses import dataclass
from enum import Enum
from typing import Optional

from ..geometry import Rectangle
from .camera import Camera
from .image import ImageSize


class ImageRotation(str, Enum):
    NONE: int = 0
    RIGHT: int = 90
    UPSIDE_DOWN: int = 180
    LEFT: int = 270


@dataclass(slots=True, kw_only=True)
class UsbCamera(Camera):
    active: bool = True
    detect: bool = False

    color: Optional[str] = None
    """a color code to identify the camera"""

    resolution: Optional[ImageSize] = None
    """physical resolution of the camera which should be used; camera may go into error state with wrong values"""

    auto_exposure: Optional[bool] = True
    """toggles auto exposure"""

    exposure: Optional[float] = None
    """manual exposure of the camera (between 0-1); set auto_exposure to False for this value to take effect"""

    rotation: ImageRotation = ImageRotation.NONE
    """rotation which should be applied after grabbing and cropping"""

    fps: Optional[int] = None
    """current frames per second (read only)"""

    crop: Optional[Rectangle] = None
    """region to crop on the original resolution before rotation"""

    @property
    def image_resolution(self) -> Optional[ImageSize]:
        if self.resolution is None:
            return None
        width = self.crop.width if self.crop else self.resolution.width
        height = self.crop.height if self.crop else self.resolution.height
        if self.rotation == ImageRotation.LEFT or self.rotation == ImageRotation.RIGHT:
            return ImageSize(width=height, height=width)
        else:
            return ImageSize(width=self.resolution.width, height=self.resolution.height)



from dataclasses import dataclass
from typing import Optional

from ..geometry import Point


@dataclass(slots=True, kw_only=True)
class Area:
    id: str
    type: Optional[str] = None
    color: str = 'green'
    outline: list[Point]



import abc
import urllib.parse
from typing import Optional
from uuid import uuid4

from .. import rosys
from ..event import Event
from .camera import Camera
from .image import Image
from .image_route import create_image_route


class CameraProvider(abc.ABC):
    """A camera provider holds a dictionary of cameras and manages additions and removals.

    The camera dictionary should not be modified directly but by using the camera provider's methods.
    This way respective events are emitted and consistency can be taken care of.

    The camera provider also creates an HTTP route to access camera images.
    """

    def __init__(self) -> None:
        self.CAMERA_ADDED = Event()
        """a new camera has been added (argument: camera)"""

        self.CAMERA_REMOVED = Event()
        """a camera has been removed (argument: camera id)"""

        self.NEW_IMAGE = Event()
        """an new image is available (argument: image)"""

        self.base_path = f'images/{str(uuid4())}'
        create_image_route(self)

    @property
    @abc.abstractmethod
    def cameras(self) -> dict[str, Camera]:
        return {}

    @property
    def images(self) -> list[Image]:
        return sorted((i for c in self.cameras.values() for i in c.images), key=lambda i: i.time)

    def get_image_url(self, image: Image) -> str:
        return f'{self.base_path}/{urllib.parse.quote_plus(image.camera_id)}/{image.time}'

    def get_latest_image_url(self, camera: Camera) -> str:
        image = camera.latest_captured_image
        if image is None:
            return f'{self.base_path}/placeholder'
        return self.get_image_url(image)

    def add_camera(self, camera: Camera) -> None:
        self.cameras[camera.id] = camera
        self.CAMERA_ADDED.emit(camera)
        self.needs_backup = True

    def remove_camera(self, camera_id: str) -> None:
        del self.cameras[camera_id]
        self.CAMERA_REMOVED.emit(camera_id)
        self.needs_backup = True

    def remove_all_cameras(self) -> None:
        [self.remove_camera(camera_id) for camera_id in list(self.cameras)]

    def remove_calibration(self, camera_id: str) -> None:
        self.cameras[camera_id].calibration = None
        self.needs_backup = True

    def remove_all_calibrations(self) -> None:
        [self.remove_calibration(camera_id) for camera_id in self.cameras]

    def add_image(self, camera: Camera, image: Image) -> None:
        camera.images.append(image)
        self.NEW_IMAGE.emit(image)

    def prune_images(self, max_age_seconds: Optional[float] = None):
        for camera in self.cameras.values():
            if max_age_seconds is None:
                camera.images.clear()
            else:
                while camera.images and camera.images[0].time < rosys.time() - max_age_seconds:
                    del camera.images[0]



from __future__ import annotations

import numpy as np

from ..geometry import Point


class Grid:

    def __init__(self, size, bbox) -> None:
        self.size = size
        self.bbox = bbox

    @staticmethod
    def from_points(points: list[Point], pixel_size: float, num_layers: int, *, padding: float = 0) -> Grid:
        min_x = min_y = np.inf
        max_x = max_y = -np.inf
        for point in points:
            min_x = min(min_x, point.x - 2.0 * pixel_size - padding)
            min_y = min(min_y, point.y - 2.0 * pixel_size - padding)
            max_x = max(max_x, point.x + 2.0 * pixel_size + padding)
            max_y = max(max_y, point.y + 2.0 * pixel_size + padding)

        width = int((max_x - min_x) / pixel_size) + 1
        height = int((max_y - min_y) / pixel_size) + 1
        return Grid((height, width, num_layers), (min_x, min_y, max_x - min_x, max_y - min_y))

    @property
    def pixel_size(self) -> float:
        h_size = self.bbox[2] / self.size[1]
        v_size = self.bbox[3] / self.size[0]
        return (h_size + v_size) / 2.0

    def to_grid(self, x, y, yaw=None) -> tuple:
        row = (y - self.bbox[1]) / self.bbox[3] * self.size[0] - 0.5
        col = (x - self.bbox[0]) / self.bbox[2] * self.size[1] - 0.5
        if yaw is None:
            return row, col

        layer = (yaw / 2.0 / np.pi * self.size[2]) % self.size[2]
        return row, col, layer

    def from_grid(self, row, col, layer=None) -> tuple:
        x = (col + 0.5) / self.size[1] * self.bbox[2] + self.bbox[0]
        y = (row + 0.5) / self.size[0] * self.bbox[3] + self.bbox[1]
        if layer is None:
            return x, y

        yaw = layer / self.size[2] * 2.0 * np.pi
        return x, y, yaw

    def contains(self, point: Point, *, padding: float = 0) -> bool:
        return self.bbox[0] + padding < point.x < self.bbox[0] + self.bbox[2] - padding \
            and self.bbox[1] + padding < point.y < self.bbox[1] + self.bbox[3] - padding



import html
import logging

import numpy as np
from nicegui import ui

from rosys import run

from .asyncio_monitor import AsyncioMonitor

log = logging.getLogger('rosys.asyncio_page')


def prepare_data(timings) -> tuple[list[str], list[dict[str, float]]]:
    def calc_box(data: list) -> dict[str, float]:
        return {
            'low': min(data),
            'q1': np.percentile(data, 25),
            'median': np.percentile(data, 50),
            'q3': np.percentile(data, 75),
            'high': max(data),
        }
    timings = dict(sorted(timings.items(), key=lambda i: len(i[1]), reverse=True))
    names = [f'{html.escape(n)} ({len(w)} warnings):<br>{html.escape(w[0].details)}' for n, w in timings.items()]
    return names, [calc_box([w.duration * 1000 for w in t]) for t in timings.values()]


class AsyncioPage(ui.page):

    def __init__(self, asyncio_monitor: AsyncioMonitor) -> None:
        super().__init__('/asyncio')

        self.asyncio_monitor = asyncio_monitor

        with self:
            with ui.row():
                ui.button('load', on_click=self.update)
                self.info_label = ui.label()
            self.chart = ui.chart(options={
                'title': False,
                'chart': {'type': 'boxplot'},
                'xAxis': {'categories': []},
                'yAxis': {'title': {'text': 'ms on UI thread'}},
                'series': {'data': []},
                'navigation': {'buttonOptions': {'enabled': False}},
                'legend': False,
                'credits': False,
            }).classes('fit').style('height:400px')

    async def update(self) -> None:
        names, data = await run.cpu_bound(prepare_data, self.asyncio_monitor.timings)
        self.chart.options['xAxis']['categories'][:] = names
        self.chart.options['series']['data'][:] = data
        self.chart.update()



from __future__ import annotations

from dataclasses import dataclass

import numpy as np


@dataclass(slots=True, kw_only=True)
class Point:
    x: float
    y: float

    @staticmethod
    def from_complex(number: complex) -> Point:
        return Point(
            x=np.real(number),
            y=np.imag(number),
        )

    @property
    def tuple(self) -> tuple[float, float]:
        return (self.x, self.y)

    @property
    def complex(self) -> complex:
        return self.x + 1j * self.y

    def distance(self, other: Point) -> float:
        return float(np.sqrt((other.x - self.x)**2 + (other.y - self.y)**2))

    def direction(self, other: Point) -> float:
        return float(np.arctan2(other.y - self.y, other.x - self.x))

    def projected_distance(self, other: Point, direction: float) -> float:
        def d(p): return np.sqrt(p.x**2 + p.y**2) * np.cos(direction - np.arctan2(p.y, p.x))
        return float(d(other) - d(self))

    def polar(self, distance: float, yaw: float) -> Point:
        return Point(
            x=self.x + distance * np.cos(yaw),
            y=self.y + distance * np.sin(yaw),
        )

    def interpolate(self, other: Point, t: float) -> Point:
        return Point(
            x=(1 - t) * self.x + t * other.x,
            y=(1 - t) * self.y + t * other.y,
        )

    def __add__(self, other: Point) -> Point:
        return Point(x=self.x + other.x, y=self.y + other.y)

    def __sub__(self, other: Point) -> Point:
        return Point(x=self.x - other.x, y=self.y - other.y)

    def __mul__(self, factor) -> Point:
        return Point(x=self.x * factor, y=self.y * factor)

    def __truediv__(self, factor) -> Point:
        return Point(x=self.x / factor, y=self.y / factor)

    def __str__(self) -> str:
        return f'Point({round(self.x, 2)}, {round(self.y, 2)})'



#!/usr/bin/env python3
import pylab as pl
from nicegui import ui
from rosys.pathplanning import plot_tools as pt
from rosys.pathplanning.experiments import generate_experiment

robot_renderer, pose, goal, obstacle_map, backward_to_goal = generate_experiment(5.0)

with ui.plot():
    pt.show_obstacle_map(obstacle_map)
    pl.gca().invert_yaxis()
    pl.autoscale(False)
    pt.plot_robot(robot_renderer, pose, 'C0')
    pt.plot_robot(robot_renderer, goal, 'C1')

ui.run()



import logging
import logging.config
import os
import sys

import coloredlogs
import icecream

import rosys


class PackagePathFilter(logging.Filter):

    # https://stackoverflow.com/a/52582536/3419103
    def filter(self, record: logging.LogRecord) -> bool:
        pathname = record.pathname
        record.relative_path = None
        abs_sys_paths = map(os.path.abspath, sys.path)
        for path in sorted(abs_sys_paths, key=len, reverse=True):  # longer paths first
            if not path.endswith(os.sep):
                path += os.sep
            if pathname.startswith(path):
                record.relative_path = os.path.relpath(pathname, path)
                break
        return True


class RosysFilter(logging.Filter):

    def filter(self, record: logging.LogRecord) -> bool:
        from rosys.test.helpers import odometer as odo
        if odo:
            record.rosys_time = rosys.time()
            record.robot_pose = f'{odo.prediction.x:.2f}, {odo.prediction.y:1.2f}, {odo.prediction.yaw_deg:1.2f}'
        else:
            record.rosys_time = 0
            record.robot_pose = 'no robot pose yet'
        return True


def setup() -> None:
    icecream.install()

    config = {
        'version': 1,
        'disable_existing_loggers': True,
        'formatters': {
            'default': {
                '()': coloredlogs.ColoredFormatter,
                'format': '%(rosys_time).2f [%(levelname)s] %(robot_pose)s %(relative_path)s:%(lineno)d: %(message)s',
                'datefmt': '%Y-%m-%d %H:%M:%S',
            },
        },
        'filters': {
            'package_path_filter': {
                '()': PackagePathFilter,
            },
            'rosys_filter': {
                '()': RosysFilter,
            }
        },
        'handlers': {
            'console': {
                'class': 'logging.StreamHandler',
                'formatter': 'default',
                'filters': ['package_path_filter', 'rosys_filter'],
                'level': 'INFO',
                'stream': 'ext://sys.stdout'
            },
        },
        'loggers': {
            '': {  # root logger
                'handlers': ['console'],
                'level': 'INFO',
                'propagate': False,
            },
            'rosys': {
                'handlers': ['console'],
                'level': 'DEBUG',
                'propagate': False,
            },
            'rosys.ota': {
                'handlers': ['console'],
                'level': 'WARN',
                'propagate': False,
            }
        },
    }

    logging.config.dictConfig(config)



from __future__ import annotations

import inspect
from dataclasses import dataclass
from typing import Callable

from .. import rosys
from ..event import Event
from ..hardware import RobotBrain
from .automator import Automator


@dataclass
class AppButton:
    icon: str
    state: str = 'enabled'
    visible: bool = True
    pressed: Callable[[], None] = lambda: None
    released: Callable[[], None] = lambda: None
    canceled: Callable[[], None] = lambda: None

    def get_properties(self) -> list[str]:
        return [
            f'/icon {self.icon}',
            f'/state {self.state}',
            f'/visible {"true" if self.visible else "false"}',
        ]

    def __str__(self) -> str:
        return f'AppButton({self.icon},{self.state},{self.visible})'


class AppControls:
    """The AppControls module enables the connection with a mobile-app-based user interface.

    It uses a given RobotBrain object to communicate with [Lizard](https://lizard.dev/) running on a microcontroller
    and in turn being connected to a mobile app via Bluetooth Low Energy.
    It displays buttons to control a given automator.
    """

    def __init__(self, robot_brain: RobotBrain, automator: Automator) -> None:
        self.APP_CONNECTED = Event()
        """an app connected via bluetooth (used to refresh information or similar)"""

        self.robot_brain = robot_brain
        self.automator = automator
        self.APP_CONNECTED

        self.main_buttons: dict[str, AppButton] = {
            'play': AppButton('play_arrow', released=automator.start),
            'pause': AppButton('pause', released=lambda: automator.pause(because='pause button was pressed')),
            'resume': AppButton('play_arrow', released=automator.resume),
            'stop': AppButton('stop', released=lambda: automator.stop(because='stop button was pressed')),
        }
        self.extra_buttons: dict[str, AppButton] = {}

        rosys.on_startup(self.sync)
        rosys.on_shutdown(self.clear)
        rosys.on_repeat(self.refresh, 0.1)
        rosys.NEW_NOTIFICATION.register(self.notify)
        robot_brain.LINE_RECEIVED.register(self.parse)

    async def refresh(self) -> None:
        before = ' '.join(str(b) for b in self.main_buttons.values())
        self.main_buttons['play'].visible = self.automator.is_stopped
        self.main_buttons['pause'].visible = self.automator.is_running
        self.main_buttons['resume'].visible = self.automator.is_paused
        self.main_buttons['play'].state = 'disabled' if self.automator.default_automation is None or not self.automator.enabled else 'enabled'
        self.main_buttons['stop'].state = 'disabled' if self.automator.is_stopped else 'enabled'
        after = ' '.join(str(b) for b in self.main_buttons.values())
        if after != before:
            await self.sync()

    async def set_info(self, msg: str) -> None:
        """replace constantly shown info text on mobile device"""
        await self.robot_brain.send(f'bluetooth.send("PUT /info {msg}")')

    async def notify(self, msg: str) -> None:
        """show notification as Snackbar message on mobile device"""
        await self.robot_brain.send(f'bluetooth.send("POST /notification {msg}")')

    def parse(self, line: str) -> None:
        if line.startswith('"'):
            line = line[1:-1]
        if line.startswith('app: '):
            line = line[5:]
            if line == 'connected':
                rosys.background_tasks.create(self.sync(), name='sync app')
                self.APP_CONNECTED.emit()
            elif line.startswith('PUT /button/') and '/action' in line:
                # line: "PUT /button/main/my_button/action pressed"
                _, path, action = line.split(' ')
                _, group, name, _ = path.split('/')
                buttons = self.main_buttons if group == 'main' else self.extra_buttons
                if action == 'pressed':
                    self._invoke(buttons[name].pressed)
                if action == 'released':
                    self._invoke(buttons[name].released)
                if action == 'canceled':
                    self._invoke(buttons[name].canceled)

    async def sync(self):
        await self.send('PUT', lambda b: b.get_properties())

    async def clear(self):
        await self.send('DELETE')
        self.main_buttons.clear()

    async def send(self, method: str, get_properties: Callable[[AppButton], list[str]] = lambda _: ['']):
        async def run(group: str, buttons: dict[str, AppButton]) -> None:
            for name, button in buttons.items():
                for prop in get_properties(button):
                    cmd = f'bluetooth.send("{method} /button/{group}/{name}{prop}")'
                    await self.robot_brain.send(cmd)
        await run('main', self.main_buttons)
        await run('extra', self.extra_buttons)

    async def _invoke(self, callback: Callable):
        if inspect.iscoroutinefunction(callback):
            rosys.background_tasks.create(callback(), name='app_controls._invoke')
        else:
            callback()



from nicegui import background_tasks

from . import analysis, automation, driving, event, geometry, hardware, pathplanning, persistence, run, system, vision
from .config import Config
from .rosys import (NEW_NOTIFICATION, Notification, config, is_test, notify, on_repeat, on_shutdown, on_startup,
                    reset_after_test, reset_before_test, set_time, shutdown, sleep, startup, time, uptime)
from .simulation_ui import simulation_ui



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['rosys',
 'rosys.analysis',
 'rosys.analysis.legacy',
 'rosys.automation',
 'rosys.driving',
 'rosys.geometry',
 'rosys.hardware',
 'rosys.hardware.communication',
 'rosys.pathplanning',
 'rosys.pathplanning.demos',
 'rosys.system',
 'rosys.test',
 'rosys.vision']

package_data = \
{'': ['*'], 'rosys.analysis': ['assets/*']}

install_requires = \
['aenum>=3.1.5,<4.0.0',
 'aiocache>=0.11.1,<0.12.0',
 'aiohttp>=3.8.3,<4.0.0',
 'aioserial>=1.3.0,<2.0.0',
 'asyncio>=3.4.3,<4.0.0',
 'coloredlogs>=15.0.1,<16.0.0',
 'dataclasses-json>=0.5.7,<0.6.0',
 'executing>=1.0.0,<2.0.0',
 'humanize>=4.0.0,<5.0.0',
 'line-profiler>=4.0.2,<5.0.0',
 'more-itertools>=8.10.0,<9.0.0',
 'msgpack>=1.0.3,<2.0.0',
 'networkx>=2.6.2,<3.0.0',
 'nicegui>=1.2.2',
 'numpy>=1.20.1,<2.0.0',
 'objgraph>=3.5.0,<4.0.0',
 'opencv-contrib-python-headless>=4.5.4,<5.0.0',
 'opencv-python>=4.5.5,<5.0.0',
 'psutil>=5.9.0,<6.0.0',
 'pyloot>=0.1.0,<0.2.0',
 'pyserial>=3.5,<4.0',
 'requests>=2.25.1,<3.0.0',
 'retry>=0.9.2,<0.10.0',
 'scipy>=1.7.2,<2.0.0',
 'sh>=1.14.2,<2.0.0',
 'simplejson>=3.17.2,<4.0.0',
 'suntime>=1.2.5,<2.0.0',
 'tabulate>=0.8.9,<0.9.0',
 'ujson==5.4.0',
 'uvloop>=0.17.0,<0.18.0',
 'yappi>=1.4,<2.0']

setup_kwargs = {
    'name': 'rosys',
    'version': '0.8.1',
    'description': 'Modular Robot System With Elegant Automation Capabilities',
    'long_description': '# RoSys - The Robot System\n\nRoSys provides an easy-to-use robot system.\nIts purpose is similar to [ROS](https://www.ros.org/).\nBut RoSys is fully based on modern web technologies and focusses on mobile robotics.\n\nThe full documentation is available at [rosys.io](https://rosys.io/).\n\n## Principles\n\n### All Python\n\nPython is great to write business logic.\nComputation-heavy tasks are wrapped in processes, accessed through WebSockets or called via C++ bindings.\nLike you would do in any other Python program.\n\n### Modularity\n\nYou can structure your code as you please.\nRoSys provides its magic without assuming a specific file structure, configuration files or enforced naming.\n\n### Event Loop\n\nThanks to [asyncio](https://docs.python.org/3/library/asyncio.html) you can write your business logic without locks and mutexes.\nThe execution is [parallel but not concurrent](https://realpython.com/python-concurrency/) which makes it easier to read, write and debug.\nIn real-case scenarios this is also much faster than [ROS](https://www.ros.org/).\nIts multiprocessing architecture requires too much inter-process communication.\n\n### Web UI\n\nMost machines need some kind of human interaction.\nRoSys is built from the ground up to make sure your robot can be operated fully off the grid with any web browser.\nThis is done by incorporating [NiceGUI](https://nicegui.io/), a wonderful all-Python UI web framework.\nIt is also possible to proxy the user interface through a gateway for remote operation.\n\n### Simulation\n\nRobot hardware is often slower than your own computer.\nTo rapidly test out new behavior and algorithms, RoSys provides a simulation mode.\nHere, all hardware is mocked and can even be manipulated to test wheel blockages and similar.\n\n### Testing\n\nYou can use [pytest](https://docs.pytest.org/) to write high-level integration tests.\nIt is based on the above-described simulation mode and accelerates the robot\'s time for super fast execution.\n\n## Architecture and Features\n\n### Modules\n\nRoSys modules are just Python modules which encapsulate certain functionality.\nThey can hold their own state, register lifecycle hooks, run methods repeatedly and subscribe to or raise [events](#events).\nModules can depend on other modules which is mostly implemented by passing them into the constructor.\n\n### Lifecycle Hooks and Loops\n\nModules can register functions for being called `on_startup` or `on_shutdown` as well as repeatedly with a given interval with `on_repeat`.\n\n### Events\n\nModules can provide events to allow connecting otherwise separated modules of the system.\nFor example, one module might read sensor data and raise an event `NEW_SENSOR_DATA`, without knowing of any consumers.\nAnother module can register on `NEW_SENSOR_DATA` and act accordingly when being called.\n\n### Automations\n\nRoSys provides an `Automator` module for running "automations".\nAutomations are coroutines that can not only be started and stopped, but also paused and resumed, e.g. using `AutomationControls`.\nHave a look at our [Click-and-drive](examples/click-and-drive.md) example.\n\n### Persistence\n\nModules can register backup and restore methods to read and write their state to disk.\n\n### Time\n\nRoSys uses its own time which is accessible through `rosys.time`.\nThis way the time can advance much faster in simulation and tests if no CPU-intensive operation is performed.\nTo delay the execution of a coroutine, you should invoke `await rosys.sleep(seconds: float)`.\nThis creates a delay until the provided amount of _RoSys time_ has elapsed.\n\n### Threading and Multiprocessing\n\nRoSys makes extensive use of [async/await](#async) to achieve parallelism without threading or multiprocessing.\nBut not every piece of code you want to integrate is offering an asyncio interface.\nTherefore RoSys provides two handy wrappers:\n\nIO-bound:\nIf you need to read from an external device or use a non-async HTTP library like [requests](https://requests.readthedocs.io/),\nyou should wrap the code in a function and await it with `await rosys.run.io_bound(...)`.\n\nCPU-bound:\nIf you need to do some heavy computation and want to spawn another process,\nyou should wrap the code in a function and await it with `await rosys.run.cpu_bound(...)`.\n\n### Safety\n\nPython (and Linux) is fast enough for most high-level logic, but has no realtime guarantees.\nSafety-relevant behavior should therefore be put on a suitable microcontroller.\nIt governs the hardware of the robot and must be able to perform safety actions like triggering emergency hold etc.\n\nWe suggest to use an industrial PC with an integrated controller like the [Zauberzeug Robot Brain](https://www.zauberzeug.com/robot-brain.html).\nIt provides a Linux system to run RoSys, offers AI acceleration via NVidia Jetson, two integrated [ESP32](https://www.espressif.com/en/products/socs/esp32) microcontrollers and six I/O sockets with up to 24 GPIOs for digital I/Os, CAN, RS485, SPI, I2C, etc.\nIt also has two hardware ENABLE switches and one which is controllable via software.\n\nTo have flexible configuration for the microcontroller we created another open source project called [Lizard](https://lizard.dev/).\nIt is a domain-specific language interpreted by the microcontroller which enables you to write reactive hardware behavior without recompiling and flashing.\n\n### User Interface\n\nRoSys builds upon the open source project [NiceGUI](https://nicegui.io/) and offers many robot-related UI elements.\nNiceGUI is a high-level UI framework for the web.\nThis means you can write all UI code in Python and the state is automatically reflected in the browser through WebSockets.\nSee any of our [examples](examples/steering.md).\n\nRoSys can also be used with other user interfaces or interaction models if required, for example a completely app-based control through Bluetooth Low Energy with Flutter.\n\n### Notifications\n\nModules can notify the user through `rosys.notify(\'message to the user\')`.\nWhen using NiceGUI, the notifications will show as snackbar messages.\nThe history of notifications is stored in the list `rosys.notifications`.\n',
    'author': 'Zauberzeug GmbH',
    'author_email': 'info@zauberzeug.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/zauberzeug/rosys',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<3.12',
}


setup(**setup_kwargs)



from dataclasses import dataclass, field
from typing import Optional
from uuid import uuid4

import numpy as np

from .. import rosys
from ..geometry import Point3d
from .camera_provider import CameraProvider
from .detections import BoxDetection, Detections, PointDetection
from .detector import Autoupload, Detector
from .image import Image
from .uploads import Uploads


@dataclass(slots=True, kw_only=True)
class SimulatedObject:
    category_name: str
    position: Point3d
    size: Optional[tuple[float]] = None
    uuid: str = field(init=False)

    def __post_init__(self) -> None:
        self.uuid = str(uuid4())


class DetectorSimulation(Detector):
    """This detector simulates object detection.

    It requires a camera provider in order to check visibility using the cameras' calibrations.
    Individual camera IDs can be added to a set of `blocked_cameras` to simulate occlusions during pytests.
    A list of `simulated_objects` can be filled to define what can be detected.
    An optional `noise` parameter controls the spatial accuracy in pixels.
    """

    def __init__(self, camera_provider: CameraProvider, *, noise: float = 1.0) -> None:
        super().__init__()

        self.camera_provider = camera_provider
        self.noise = noise

        self.blocked_cameras: set[str] = set()
        self.simulated_objects: list[SimulatedObject] = []
        self._uploads = Uploads()

        rosys.on_repeat(self.step, 0.1)

    @property
    def uploads(self) -> Uploads:
        return self._uploads

    def step(self) -> None:
        self._uploads.queue.clear()
        self._uploads.priority_queue.clear()

    async def detect(self, image: Image, autoupload: Autoupload = Autoupload.FILTERED) -> Optional[Detections]:
        is_blocked = image.camera_id in self.blocked_cameras
        await rosys.sleep(0.4)
        image.detections = Detections()
        if not is_blocked:
            self.update_simulated_objects(image)
            self.detect_from_simulated_objects(image)
        self.NEW_DETECTIONS.emit(image)
        return image.detections

    async def upload(self, image: Image) -> None:
        self.log.info(f'Uploading {image.id}')
        await super().upload(image)

    def update_simulated_objects(self, image: Image) -> None:
        pass

    def detect_from_simulated_objects(self, image: Image) -> None:
        if image.camera_id not in self.camera_provider.cameras:
            return
        camera = self.camera_provider.cameras[image.camera_id]
        for object in self.simulated_objects:
            viewing_direction = np.array(camera.calibration.extrinsics.rotation.R)[:, 2]
            object_direction = np.array(object.position.tuple) - camera.calibration.extrinsics.translation
            if np.dot(viewing_direction, object_direction) < 0:
                continue
            image_point = camera.calibration.project_to_image(object.position)
            if not (0 <= image_point.x < image.size.width and 0 <= image_point.y < image.size.height):
                continue

            if object.size is None:
                image.detections.points.append(PointDetection(
                    category_name=object.category_name,
                    model_name='simulation',
                    confidence=1.0,
                    x=image_point.x + self.noise * np.random.randn(),
                    y=image_point.y + self.noise * np.random.randn(),
                    uuid=object.uuid,
                ))
            else:
                world_points = np.array([
                    [object.position.x + dx, object.position.y + dy, object.position.z + dz]
                    for dx in [-object.size[0] / 2, object.size[0] / 2]
                    for dy in [-object.size[1] / 2, object.size[1] / 2]
                    for dz in [-object.size[2] / 2, object.size[2] / 2]
                ])
                image_points = camera.calibration.project_array_to_image(world_points)
                image.detections.boxes.append(BoxDetection(
                    category_name=object.category_name,
                    model_name='simulation',
                    confidence=1.0,
                    x=image_points[:, 0].min() + self.noise * np.random.randn(),
                    y=image_points[:, 1].min() + self.noise * np.random.randn(),
                    width=image_points[:, 0].max() - image_points[:, 0].min() + self.noise * np.random.randn(),
                    height=image_points[:, 1].max() - image_points[:, 1].min() + self.noise * np.random.randn(),
                    uuid=object.uuid,
                ))



import logging

import numpy as np
from nicegui import ui
from nicegui.events import KeyEventArguments

from ..geometry import Point
from .steerer import Steerer


class KeyboardControl:
    """The KeyboardControl UI element allows controlling a given steerer via keyboard events.

    Hold shift while pressing an arrow key to steer the robot.
    You can change the speed with the number keys 1 to 9 and the initial speed via the `default_speed` argument.
    """

    def __init__(self, steerer: Steerer, *, default_speed: float = 2.0) -> None:
        self.steerer = steerer
        self.log = logging.getLogger('rosys.ui.keyboard_control')
        ui.keyboard(on_key=self.handle_keys, repeating=False)
        self.direction = Point(x=0, y=0)
        self.speed = default_speed

    def handle_keys(self, e: KeyEventArguments) -> None:
        self.log.debug(f'{e.key.name} -> {e.action} {e.modifiers}')

        # change speed via number key
        if e.action.keydown and e.key.number is not None:
            self.speed = e.key.number
            self.direction.y = self.speed * np.sign(self.direction.y)
            self.direction.x = self.speed * np.sign(self.direction.x)
            self.steerer.update(self.direction.y, self.direction.x)

        # update individual speed components via arrow keys
        if e.modifiers.shift and e.key.is_cursorkey:
            if e.action.keydown and self.direction.x == 0 and self.direction.y == 0:
                self.steerer.start()

            new_speed = self.speed if e.action.keydown else 0
            if e.key.arrow_left:
                self.direction.y = -new_speed
            elif e.key.arrow_right:
                self.direction.y = +new_speed
            elif e.key.arrow_up:
                self.direction.x = +new_speed
            elif e.key.arrow_down:
                self.direction.x = -new_speed
            self.steerer.update(self.direction.y, self.direction.x)

            if e.action.keyup and self.direction.x == 0 and self.direction.y == 0:
                self.steerer.stop()

        # stop when releasing shift key
        if e.key.shift and e.action.keyup:
            self.direction.y = 0
            self.direction.x = 0
            self.steerer.stop()



from __future__ import annotations

import numpy as np
from nicegui import ui
from nicegui.elements.scene_objects import Group, Sphere

from ..rosys import config
from .driver import Driver


class DriverObject(Group):
    """The DriverObject UI element displays the path following process in a 3D scene.

    The current pose is taken from a given odometer.
    An optional driver module shows debugging information about a current path-following process.
    The `debug` argument can be set to show a wireframe instead of a closed polygon.
    """

    def __init__(self, driver: Driver) -> None:
        super().__init__()
        self.driver = driver
        with self:
            with Group() as self.robot_group:
                Sphere(0.03).material('#4488ff')
                self.hook = Sphere(0.05).material('#4488ff')
                self.pull = Sphere(0.05).material('#ffbb00')
            with Group() as self.carrot_group:
                Sphere(0.03).material('#ff8800')
                Sphere(0.05).material('#ff8800').move(self.driver.parameters.carrot_offset)
        ui.timer(config.ui_update_interval, self.update)

    def update(self) -> None:
        robot_pose = self.driver.odometer.prediction
        self.robot_group.move(robot_pose.x, robot_pose.y).rotate(0, 0, robot_pose.yaw)

        if self.driver.state:
            self.scale(1)
            carrot_pose = self.driver.state.carrot_pose
            self.carrot_group.move(carrot_pose.x, carrot_pose.y).rotate(0, 0, carrot_pose.yaw)
            angle = self.driver.state.turn_angle + (np.pi if self.driver.state.backward else 0)
            hook_offset = self.driver.parameters.hook_offset * (-1 if self.driver.state.backward else 1)
            pull_distance = 1.5 * self.driver.parameters.carrot_distance
            self.hook.move().move(hook_offset)
            self.pull.move(hook_offset + np.cos(angle) * pull_distance, np.sin(angle) * pull_distance)
        else:
            self.scale(0)



from .line import Line
from .line_segment import LineSegment
from .point import Point
from .point3d import Point3d
from .pose import Pose, PoseStep
from .prism import Prism
from .rectangle import Rectangle
from .rotation import Rotation
from .spline import Spline
from .velocity import Velocity
